# -*- coding: utf-8 -*-
# These classes are a simplified version of Disease and Diseases classes
# originally developed by Orion Buske in patient-similarity (https://github.com/buske/patient-similarity)

import logging
import re
from collections import defaultdict

from patientMatcher.resources import path_to_phenotype_annotations

LOG = logging.getLogger(__name__)
db_re = re.compile(r"([A-Z]+:\d+)")

FREQUENCY_TERMS = {
    "HP:0040280": 1.0,  # Obligate
    "HP:0040281": (0.99 + 0.80) / 2.0,  # Very frequent
    "HP:0040282": (0.79 + 0.30) / 2.0,  # Frequent
    "HP:0040283": (0.05 + 0.29) / 2.0,  # Occasional
    "HP:0040284": (0.01 + 0.04) / 2.0,  # Very rare
    "HP:0040285": 0.0,  # Excluded
}

FREQUENCIES = {
    "very rare": 0.01,
    "rare": 0.05,
    "occasional": 0.075,
    "frequent": 0.33,
    "typical": 0.5,
    "variable": 0.5,
    "common": 0.75,
    "hallmark": 0.9,
    "obligate": 1.0,
}


class Disease:
    """An object representing a single disease"""

    def __init__(self, db, db_id, phenotypes):
        self.db = db
        self.id = db_id
        self.phenotype_freqs = phenotypes

    def __str__(self):
        return f"{self.db}:{self.id}"


class Diseases:
    """Create an object containing all diseases from the phenotype_annotations.tav.txt file
    Resources included in this file: DECIPHER, OMIM, ORPHANET
    """

    def init_app(self, app):
        """Initialize the diseases object when the app is launched"""
        self.databases = ["DECIPHER", "OMIM", "ORPHA"]
        self.diseases = {}
        self._parse_diseases()
        LOG.info(f"Parsed {len(self.diseases.keys())} disease/phenotypes from resource file")

    def _parse_disease_frequency(self, field):
        """Parse disease frequency (col 8 in phenotype anno file)"""
        if not field:
            return None

        if field.upper() in FREQUENCY_TERMS:
            return FREQUENCY_TERMS[field]

        field = field.lower()

        if field in FREQUENCIES:
            return FREQUENCIES[field]
        if field.endswith("%"):
            field = field.replace("%", "")
            if "-" in field:
                # Average any frequency ranges
                low, high = field.split("-")
                freq = (float(low) + float(high)) / 2 / 100
            else:
                freq = float(field) / 100
        else:
            try:
                num, denom = fraction_frequency_re.split(field)
            except Exception as ex:
                LOG.warning(f"Error parsing frequency: {field} -> {ex}")
                freq = default
            else:
                freq = float(num) / float(denom)
        return freq

    def _parse_diseases(self):
        """Parse diseases file, that is available under patientMatcher/resources"""
        disease_phenotypes = defaultdict(dict)
        with open(path_to_phenotype_annotations, encoding="utf-8") as disease_file:
            for line in disease_file:
                diseases = []
                line = line.strip()
                items = line.split("\t")

                db = items[0]
                if db not in self.databases:
                    continue
                diseases.append((items[0].strip(), items[1].strip()))  # diseases: [(OMIM, 102400)]

                # Add alternative terms to list of diseases
                alt_diseases = db_re.findall(items[5].strip())  # example: ['OMIM:147791']
                for alt_disease in alt_diseases:
                    alt_db = alt_disease.split(":")[0].strip()
                    alt_db_id = int(alt_disease.split(":")[1].strip())
                    if alt_db in ["MIM", "IM"]:
                        alt_db = "OMIM"
                    if alt_db not in self.databases:
                        continue
                    diseases.append((alt_db, alt_db_id))

                # Add HPO terms and frequencies to disease terms
                hpo_term = items[4].strip()
                freq = self._parse_disease_frequency(items[8])
                for disease in diseases:
                    phenotypes = disease_phenotypes[disease]
                    # Collect max annotated frequency
                    if freq is not None and hpo_term in phenotypes:
                        old_freq = phenotypes[hpo_term]
                        if old_freq is None or old_freq < freq:
                            phenotypes[hpo_term] = freq
                        if old_freq != freq:
                            pass
                    else:
                        phenotypes[hpo_term] = freq
                    # phenotypes[hpo_term] = None  # Do not populate disease frequencies

        for (db, db_id), phenotypes in disease_phenotypes.items():
            disease = Disease(db, db_id, phenotypes)
            self.diseases[(db, db_id)] = disease



# -*- coding: utf-8 -*-
import patientMatcher.utils.ensembl_rest_client as ensembl_client


def liftover(build, chrom, start, end=None):
    """Perform variant liftover using Ensembl REST API

    Accepts:
        build(str): genome build: GRCh37 or GRCh38
        chrom(str): 1-22,X,Y,MT
        start(int): start coordinate
        stop(int): stop coordinate or None

    Returns
        mappings(list of dict):
            example: https://rest.ensembl.org/map/human/GRCh37/X:1000000..1000100:1/GRCh38?content-type=application/json
    """
    assembly2 = "GRCh38"
    if build == "GRCh38":
        assembly2 = "GRCh37"

    client = ensembl_client.EnsemblRestApiClient()
    url = "/".join(
        [
            client.server,
            "map/human",
            build,
            f"{chrom}:{start}..{end or start}",  # End variant provided is not required
            f"{assembly2}?content-type=application/json",
        ]
    )
    result = client.send_request(url)
    if isinstance(result, dict):
        return result.get("mappings")



# -*- coding: utf-8 -*-

import logging

LOG = logging.getLogger(__name__)


def authorize(database, request):
    """Validate request's token against database client collection

    Args:
        database(pymongo.database.Database)
        request(request): a request object

    Returns:
        authorized(bool): True or False
    """
    token = request.headers.get("X-Auth-Token")
    query = {"auth_token": token}
    authorized = database["clients"].find_one(query)
    if bool(authorized):
        LOG.info(
            'Authorized client with id "{0}" submits a {1} request'.format(
                authorized["_id"], request.method
            )
        )
    return bool(authorized)



# -*- coding: utf-8 -*-

from patientMatcher.match.phenotype_matcher import match, similarity_wrapper
from patientMatcher.parse.patient import mme_patient
from patientMatcher.resources import path_to_hpo_terms, path_to_phenotype_annotations
from patientMatcher.server.extensions import diseases, hpo, hpoic
from patientMatcher.utils.patient import Patient

PHENOTYPE_ROOT = "HP:0000118"


def test_patient_similarity_wrapper(mock_app):
    """Test the function that calculates the HPO similarity between 2 patients"""

    assert hpo
    assert diseases
    assert hpoic

    query_p_terms = [
        "HP:0008058",
        "HP:0007033",
        "HP:0002194",
        "HP:0002281",
    ]  # nervous system - related HPO terms

    # test wrapper by providing same terms for query patient and match patient:
    score = similarity_wrapper(
        hpoic=hpoic,
        hpo=hpo,
        max_hpo_score=1.0,
        hpo_terms_q=query_p_terms,
        hpo_terms_m=query_p_terms,
    )
    # score should be something like 0.9999999999999999
    assert round(score, 12) == 1

    match_p_terms = ["HP:0008058", "HP:0007033", "HP:0002194"]

    # test wrapper by providing almost the same terms for query patient and match patient:
    related_pheno_score = similarity_wrapper(
        hpoic=hpoic,
        hpo=hpo,
        max_hpo_score=1.0,
        hpo_terms_q=query_p_terms,
        hpo_terms_m=match_p_terms,
    )
    # similarity score should be lower, but still around 1
    assert round(related_pheno_score, 2) < score
    assert related_pheno_score > 0.8

    # provide completely different HPO terms for matching patient
    match_p_terms = ["HP:0003002", "HP:0000218"]  # breast cancer and high palate phenotypes
    unrelated_pheno_score = similarity_wrapper(
        hpoic=hpoic,
        hpo=hpo,
        max_hpo_score=1.0,
        hpo_terms_q=query_p_terms,
        hpo_terms_m=match_p_terms,
    )
    # then unrelated_pheno_score should be almost 0
    assert round(unrelated_pheno_score, 2) == 0

    # but still a positive number
    assert unrelated_pheno_score > 0


def test_phenotype_matching(gpx4_patients, database):
    """test the algorithm that compares the phenotype of a query patient against the database"""

    # insert 2 test patients into test database
    for patient in gpx4_patients:
        database["patients"].insert_one(patient)
    assert len(list(database["patients"].find())) == 2

    query_patient = gpx4_patients[0]
    assert query_patient

    # this patient has HPO terms and OMIM diagnosis
    formatted_patient = mme_patient(query_patient)
    assert len(formatted_patient["features"]) > 0
    assert len(formatted_patient["disorders"]) > 0

    matches_HPO_OMIM = match(
        database, 0.75, formatted_patient["features"], formatted_patient["disorders"]
    )
    assert len(matches_HPO_OMIM.keys()) == 2
    for key, value in matches_HPO_OMIM.items():
        assert "patient_obj" in value
        assert value["pheno_score"] > 0

    features = formatted_patient["features"]
    disorders = formatted_patient["disorders"]
    # remove HPO terms from the query patient, test that the algorithm works anyway
    # because matching will use OMIM disorders
    formatted_patient["features"] = []
    matches_OMIM = match(
        database, 0.75, formatted_patient["features"], formatted_patient["disorders"]
    )
    assert len(matches_OMIM.keys()) > 0 and len(matches_OMIM.keys()) < 50
    for key, value in matches_OMIM.items():
        assert "patient_obj" in value
        assert value["pheno_score"] > 0

    # remove the OMIM diagnosis from patient object. The algorithm should work
    # but it shouldn't return any match
    formatted_patient["disorders"] = []
    matches_no_phenotypes = match(
        database, 0.75, formatted_patient["features"], formatted_patient["disorders"]
    )
    assert len(matches_no_phenotypes.keys()) == 0

    # Add again features. The algorithm works again because HPO terms will be used
    formatted_patient["features"] = features
    matches_HPO = match(
        database, 0.75, formatted_patient["features"], formatted_patient["disorders"]
    )
    assert len(matches_HPO.keys()) == 2
    for key, value in matches_HPO.items():
        assert "patient_obj" in value
        assert value["pheno_score"] > 0

    # make sure that matches obtained when OMIM and HPO terms are present are more or equal than
    # when either of these phenotype terms is present by itself
    assert len(matches_HPO_OMIM.keys()) >= len(matches_OMIM.keys())
    assert len(matches_HPO_OMIM.keys()) >= len(matches_HPO.keys())



# -*- coding: utf-8 -*-
import os

from patientMatcher.resources import path_to_hpo_terms
from patientMatcher.server import configure_email_error_logging, create_app
from patientMatcher.utils.notify import TlsSMTPHandler


def test_create_app():
    """Tests the function that creates the app"""

    assert create_app()


def test_create_app_missing_requirement(monkeypatch):
    """Tests the function that creates the app when phenotype_annotation.tab.txt and hp.obo.txt
    resources are missing"""

    # GIVEN a prod server
    monkeypatch.setenv("TESTING", False)

    # GIVEN one of the required files missing / it's been renamed
    temp_file = ".".join([path_to_hpo_terms, "temp"])
    os.rename(path_to_hpo_terms, temp_file)

    # THEN the app should not be created
    app = create_app()
    assert app is None

    # Revert original resource name
    os.rename(temp_file, path_to_hpo_terms)


def test_error_log_email(mock_app):
    """Test the app error logging via email"""

    # GIVEN an app with an ADMIN and configured email error logging params
    mail_host = "smtp.gmail.com"
    mail_port = 587
    server_email = "server_email"
    server_pw = "server_pw"

    mock_app.config["ADMINS"] = ["app_admin_email"]
    mock_app.config["MAIL_SERVER"] = mail_host
    mock_app.config["MAIL_PORT"] = mail_port
    mock_app.config["MAIL_USERNAME"] = server_email
    mock_app.config["MAIL_PASSWORD"] = server_pw

    configure_email_error_logging(mock_app)

    # Then a TlsSMTPHandler should be among the app loggers
    handler = mock_app.logger.handlers[0]
    assert isinstance(handler, TlsSMTPHandler)
    # And should contain the given settings
    assert handler.mailhost == mail_host
    assert handler.mailport == mail_port
    assert handler.fromaddr == server_email
    assert handler.password == server_pw
    assert handler.toaddrs == mock_app.config["ADMINS"]



# -*- coding: UTF-8 -*-
import logging

import requests

LOG = logging.getLogger(__name__)

HEADERS = {"Content-type": "application/json"}
RESTAPI_37 = "https://grch37.rest.ensembl.org"
RESTAPI_38 = "https://rest.ensembl.org/"
PING_ENDPOINT = "info/ping"


class EnsemblRestApiClient:
    """A class handling requests and responses to and from the Ensembl REST APIs.
    Endpoints for human build 37: https://grch37.rest.ensembl.org
    Endpoints for human build 38: http://rest.ensembl.org/
    Documentation: https://github.com/Ensembl/ensembl-rest/wiki
    doi:10.1093/bioinformatics/btu613
    """

    def __init__(self, build="37"):
        if build == "38":
            self.server = RESTAPI_38
        else:
            self.server = RESTAPI_37

    def ping_server(self, server=RESTAPI_38):
        """ping ensembl

        Accepts:
            server(str): default is 'https://grch37.rest.ensembl.org'

        Returns:
            data(dict): dictionary from json response
        """
        url = "/".join([server, PING_ENDPOINT])
        data = self.send_request(url)
        return data

    def except_on_invalid_response(self, resp):
        """Checks if Ensembl service returned a valid response(status:200->OK), otherwise raise error with informative message"""
        # If service returned error
        if resp.status != 200:
            # raise an exception with a proper message
            raise Exception(
                f"Ensembl Rest API server returned error {str(resp.status)} (it's probably down) and gene info could not be converted. Please try again later."
            )

    def send_request(self, url):
        """Sends the actual request to the server and returns the response

        Accepts:
            url(str): ex. https://grch37.rest.ensembl.org/lookup/id/ENSG00000103591

        Returns:
            data(dict): dictionary from json response
        """
        data = {}
        try:
            response = requests.get(
                url,
                headers=HEADERS,
            )
            data = response.json()
            if response.status_code != 200:
                raise ValueError("The API did not return valid data")
        except requests.exceptions.MissingSchema as ex:
            LOG.error("Request failed for url {0}: Missing Schrma error: {1}\n".format(url, ex))
            data = ex
        except ValueError as ex:
            LOG.error("Request failed for url {0}. Value Error: {1}\n".format(url, ex))
            data = ex

        return data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from patientMatcher.parse import patient


def test_href_validate_wrong_url():
    """Test href_validate function with a malformed URL"""
    assert patient.href_validate("google") is False


def test_href_validate_valid_url():
    """Test href_validate function with a valid URL"""
    assert patient.href_validate("http://google.com") is True


def test_href_validate_wrong_email():
    """Test href_validate function with a mailto link and wrong email syntax"""
    assert patient.href_validate("me@patientmatcher.se") is False


def test_href_validate_valid_email():
    """Test href_validate function with a mailto link and valid email"""
    assert patient.href_validate("mailto:me@patientmatcher.se") is True


def test_features_to_hpo_no_features():
    """Make sure the function returns [] if patient doesn't have associated HPO terms"""
    result = patient.features_to_hpo(None)
    assert result == []


def test_disorders_to_omim_no_omim():
    """Make sure the function returns [] if patient doesn't have associated OMIM terms"""
    result = patient.disorders_to_omim(None)
    assert result == []


def test_mme_patient_gene_symbol(gpx4_patients, monkeypatch):
    """Test format a patient with HGNC gene symbol"""

    # GIVEN a patched gene conversion system mocking the Ensembl web services
    def mock_symbol_to_ensembl(*args):
        return "ENSG00000167468"

    monkeypatch.setattr(patient, "symbol_to_ensembl", mock_symbol_to_ensembl)

    test_patient = gpx4_patients[0]
    gene_name = test_patient["genomicFeatures"][0]["gene"]["id"]  # "GPX4"
    # Before conversion patient's gene id is a gene symbol
    assert gene_name.startswith("ENSG") is False
    mme_formatted_patient = patient.mme_patient(
        test_patient, True
    )  # Convert gene symbol to Ensembl
    # After conversion formatted patient's gene id should be an Ensembl id
    assert mme_formatted_patient["genomicFeatures"][0]["gene"]["id"].startswith("ENSG")
    assert mme_formatted_patient["genomicFeatures"][0]["gene"]["_geneName"] == gene_name


def test_mme_patient_entrez_gene(entrez_gene_patient, monkeypatch):
    """Test format a patient with entrez gene"""

    # GIVEN a patched gene conversion system mocking the Ensembl web services
    def mock_symbol_to_ensembl(*args):
        return "ENSG00000167468"

    def mock_entrez_to_symbol(*args):
        return "KARS"

    monkeypatch.setattr(patient, "symbol_to_ensembl", mock_symbol_to_ensembl)
    monkeypatch.setattr(patient, "entrez_to_symbol", mock_entrez_to_symbol)

    # Before conversion patient's gene id is an entrez gene ID
    assert entrez_gene_patient["genomicFeatures"][0]["gene"]["id"] == "3735"
    mme_formatted_patient = patient.mme_patient(
        entrez_gene_patient, True
    )  # convert genes to Ensembl
    # After conversion formatted patient's gene id should be an Ensembl id
    assert mme_formatted_patient["genomicFeatures"][0]["gene"]["id"].startswith("ENSG")
    assert mme_formatted_patient["genomicFeatures"][0]["gene"]["_geneName"]  # it's "KARS"


def test_gtfeatures_to_variants(patient_37, monkeypatch):
    """Test the function that parses variants dictionaries from patient's genomic features"""

    # GIVEN a mocked Ensembl REST API liftover service
    def mock_liftover(*args):
        mapped = {
            "mapped": {
                "assembly": "GRCh38",
                "seq_region_name": "12",
                "start": 14641141,
                "end": 14641142,
            }
        }
        return [mapped]

    monkeypatch.setattr(patient, "liftover", mock_liftover)

    # GIVEN a patient containing 1 genomic feature (and one variant)
    gt_features = patient_37["patient"]["genomicFeatures"]
    assert len(gt_features) == 1

    # WHEN gtfeatures_to_variants is used to extract variants from gt_features
    variants = patient.gtfeatures_to_variants(gt_features)

    # THEN it should return 2 variants
    assert len(variants) == 2

    # One with genome build GRCh37
    assert variants[0]["assembly"] == "GRCh37"
    # And one with genome build GRCh38
    assert variants[1]["assembly"] == "GRCh38"



# -*- coding: utf-8 -*-

import json
import logging

from bson import json_util
from flask import Blueprint, current_app, jsonify, render_template, request
from flask_negotiate import consumes, produces
from patientMatcher.auth.auth import authorize
from patientMatcher.match.handler import async_match, internal_matcher, patient_matches
from patientMatcher.utils.add import backend_add_patient
from patientMatcher.utils.notify import notify_match_external, notify_match_internal

from . import controllers

LOG = logging.getLogger(__name__)
blueprint = Blueprint("server", __name__)
API_MIME_TYPE = "application/vnd.ga4gh.matchmaker.v1.0+json"


@blueprint.route("/", methods=["GET"])
def index():
    """Provide an index page with general node description and stats"""
    data = controllers.populate_index_data()
    return render_template("index.html", **data)


@blueprint.route("/patient/add", methods=["POST"])
@consumes(API_MIME_TYPE, "application/json")
@produces("application/json")
def add():
    """Add a patient to the database"""

    formatted_patient = controllers.check_request(current_app.db, request)
    if isinstance(formatted_patient, int):  # some error must have occurred during validation
        return controllers.bad_request(formatted_patient)

    match_external = False
    if controllers.get_nodes(database=current_app.db):
        match_external = True

    # else import patient to database
    modified, inserted, matching_obj = backend_add_patient(
        mongo_db=current_app.db, patient=formatted_patient, match_external=match_external
    )
    message = {}

    if modified:
        message["message"] = "Patient was successfully updated."
    elif inserted:
        message["message"] = "Patient was successfully inserted into database."
    else:
        message["message"] = "Database content is unchanged."

    # if patient is matching any other patient on other nodes
    # and match notifications are on
    if current_app.config.get("MAIL_SERVER") and matching_obj and len(matching_obj.get("results")):
        # send an email to patient's contact:
        notify_match_external(
            match_obj=matching_obj,
            admin_email=current_app.config.get("MAIL_USERNAME"),
            mail=current_app.mail,
            notify_complete=current_app.config.get("NOTIFY_COMPLETE"),
        )

    resp = jsonify(message)
    resp.status_code = 200
    return resp


@blueprint.route("/patient/delete/<patient_id>", methods=["DELETE"])
def delete(patient_id):
    """Delete a patient from the database using its ID"""
    # check if request is authorized
    resp = None
    if not authorize(current_app.db, request):  # not authorized, return a 401 status code
        return controllers.bad_request(401)

    LOG.info("Authorized client is removing patient with id {}".format(patient_id))
    message = controllers.delete_patient(current_app.db, patient_id)
    resp = jsonify(message)
    resp.status_code = 200
    return resp


@blueprint.route("/heartbeat", methods=["GET"])
def heartbeat():
    """Get the server specs"""
    resp = None
    if authorize(current_app.db, request):
        LOG.info("Authorized client requests heartbeat..")
        LOG.info("current_app is {}".format(current_app))
        disclaimer = current_app.config.get("DISCLAIMER")
        result = controllers.heartbeat(disclaimer)
        resp = jsonify(result)
        resp.status_code = 200

    else:  # not authorized, return a 401 status code
        return controllers.bad_request(401)

    return resp


@blueprint.route("/metrics", methods=["GET"])
def metrics():
    """Get database metrics"""
    LOG.info("Receiving a requests for server metrics.")
    results = controllers.metrics()
    resp = jsonify({"metrics": results, "disclaimer": current_app.config.get("DISCLAIMER")})
    resp.status_code = 200
    return resp


@blueprint.route("/nodes", methods=["GET"])
def nodes():
    """Get a list of all nodes connected to this MME server"""
    resp = None
    if not authorize(current_app.db, request):
        return controllers.bad_request(401)

    LOG.info("Authorized client requests list of connected nodes..")
    results = controllers.get_nodes(database=current_app.db)
    resp = jsonify(results)
    resp.status_code = 200
    return resp


@blueprint.route("/matches/<patient_id>", methods=["GET"])
def matches(patient_id):
    """Get all matches (external and internal) for a patient ID"""
    LOG.info("getting all matches for patient {}".format(patient_id))
    resp = None
    message = {}
    if not authorize(current_app.db, request):  # not authorized, return a 401 status code
        return controllers.bad_request(401)

    # return only matches with at least one result
    results = patient_matches(current_app.db, patient_id)
    if results:
        message = json.loads(json_util.dumps({"matches": results}))
    else:
        message["message"] = "Could not find any matches in database for patient ID {}".format(
            patient_id
        )
    resp = jsonify(message)
    resp.status_code = 200
    return resp


@blueprint.route("/match/external/<patient_id>", methods=["POST"])
def match_external(patient_id):
    """Trigger a patient matching on external nodes by providing a patient ID"""
    resp = None
    message = {}
    if not authorize(current_app.db, request):  # not authorized, return a 401 status code
        return controllers.bad_request(401)

    LOG.info(
        "Authorized clients is matching patient with ID {} against external nodes".format(
            patient_id
        )
    )
    query_patient = controllers.patient(current_app.db, patient_id)

    if not query_patient:
        LOG.error("ERROR. Could not find any patient with ID {} in database".format(patient_id))
        message["message"] = "ERROR. Could not find any patient with ID {} in database".format(
            patient_id
        )
        resp = jsonify(message)
        resp.status_code = 200
        return resp

    node = request.args.get("node")

    # if search should be performed on a specific node, make sure node is in database
    if node and current_app.db["nodes"].find_one({"_id": node}) is None:
        LOG.info('ERROR, theres no node with id "{}" in database'.format(request.args["node"]))
        message[
            "message"
        ] = "ERROR. Could not find any connected node with id {} in database".format(
            request.args["node"]
        )
        resp = jsonify(message)
        resp.status_code = 200
        return resp

    matching_obj = controllers.match_external(current_app.db, query_patient, node)

    if not matching_obj:
        message["message"] = "Could not find any other node connected to this MatchMaker server"
        resp = jsonify(message)
        resp.status_code = 200
        return resp

    results = matching_obj.get("results")

    # if patient is matching any other patient on other nodes
    # and match notifications are on
    if current_app.config.get("MAIL_SERVER") and matching_obj and len(results):
        # send an email to patient's contact:
        notify_match_external(
            match_obj=matching_obj,
            admin_email=current_app.config.get("MAIL_USERNAME"),
            mail=current_app.mail,
            notify_complete=current_app.config.get("NOTIFY_COMPLETE"),
        )

    resp = jsonify({"results": results})
    resp.status_code = 200
    return resp


@blueprint.route("/async_response", methods=["POST"])
def asynch_match_response():
    """Receives and handles asynchronous match responses.
    This implies that patientMatcher has already submitted a match request
    to a server x and has received not results, but a query id in response.
    Query id was previously saved into patientMatcher database.
    This delayed response from server x contains match results and
    the same query id and server x source key.
    """
    LOG.info("Receiving an asynchronous request.")
    data = controllers.check_async_request(current_app.db, request)
    # returned data after check_async_request is a dictionary if request is valid
    if isinstance(data, int):  # some error must have occurred during validation
        return controllers.bad_request(data)

    # create a match object from specifics you first sent to the node
    async_match_obj = async_match(current_app.db, data)
    if async_match_obj:
        # save match object to database
        LOG.info("saving match object to database")
        current_app.db["matches"].insert_one(async_match_obj)

        # Remove query_id from async responses, this is not used any more
        current_app.db["async_responses"].delete_one({"query_id": data["query_id"]})
    else:
        message = "Error: could not create a valid match object from request data"
        resp = jsonify({"message": message})
        resp.status_code = 200
        return resp

    # notify matches if there are matching patients and app is configured to do so
    if len(async_match_obj["results"]) > 0 and current_app.config.get("MAIL_SERVER"):
        # send an email to patient's contact:
        notify_match_external(
            match_obj=async_match_obj,
            admin_email=current_app.config.get("MAIL_USERNAME"),
            mail=current_app.mail,
            notify_complete=current_app.config.get("NOTIFY_COMPLETE"),
        )

    resp = jsonify({"message": "results received, many thanks!"})
    resp.status_code = 200
    return resp


@blueprint.route("/match", methods=["POST"])
@consumes(API_MIME_TYPE)
@produces(API_MIME_TYPE, "application/json")
def match_internal():
    """Match a query patient against patients from local database and
    returns a response containing patients which are most similar to the
    query patient"""

    query_patient = controllers.check_request(current_app.db, request)
    if isinstance(query_patient, int):  # some error must have occurred during validation
        return controllers.bad_request(query_patient)

    LOG.info("Matching a query patient against database")

    max_pheno_score = current_app.config.get(
        "MAX_PHENO_SCORE", 0.25
    )  # get max pheno score from app settings, if available
    max_geno_score = current_app.config.get(
        "MAX_GT_SCORE", 0.75
    )  # get max genotyping score from app settings, if available
    max_results = current_app.config.get("MAX_RESULTS", 5)
    score_threshold = current_app.config.get("SCORE_THRESHOLD", 0)

    # get a list of matching patients ordered by score
    match_obj = internal_matcher(
        current_app.db, query_patient, max_pheno_score, max_geno_score, max_results, score_threshold
    )
    # save matching object to database
    current_app.db["matches"].insert_one(match_obj)
    matches = []
    try:
        matches = match_obj["results"][0][
            "patients"
        ]  # results[0] because there is just one node (internal match)
    except Exception as ex:
        LOG.info("An error has occurred while executing an internal patient match:{}".format(ex))

    # if notifications are on and there are matching results
    if current_app.config.get("MAIL_SERVER") and len(matches):
        notify_match_internal(
            database=current_app.db,
            match_obj=match_obj,
            admin_email=current_app.config.get("MAIL_USERNAME"),
            mail=current_app.mail,
            notify_complete=current_app.config.get("NOTIFY_COMPLETE"),
        )

    # validate_response = controllers.validate_response({'results': matches})
    validate_response = {"results": matches}
    if isinstance(validate_response, int):  # some error must have occurred during validation
        return controllers.bad_request(validate_response)

    LOG.info("Returning {} matching patients from database".format(len(matches)))

    # return response with results
    validate_response["disclaimer"] = current_app.config.get("DISCLAIMER")
    resp = jsonify(validate_response)
    resp.status_code = 200
    return resp



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import datetime
import json
import logging

import requests
from patientMatcher.match.genotype_matcher import match as genomatch
from patientMatcher.match.phenotype_matcher import match as phenomatch
from patientMatcher.parse.patient import json_patient
from werkzeug.datastructures import Headers

LOG = logging.getLogger(__name__)


def patient_matches(database, patient_id, type=None, with_results=True):
    """Retrieve all matches for a patient specified by an ID

    Args:
        database(pymongo.database.Database)
        patient_id(str): ID of a patient in database
        type(str): type of matching, external or internal
        with_results(bool): if True only matches with results are returned

    Returns:
        matches(list): a list of dictionaries = [ {match_obj1}, {match_obj2}, .. ]
    """
    query = {
        "$or": [
            {"data.patient.id": patient_id},  # collect matches triggered by patient
            {
                "results.patients.patient.id": patient_id
            },  # and matches where patient is among results
        ]
    }
    if type:
        query["match_type"] = type
    if with_results:
        query["has_matches"] = True

    matches = list(database["matches"].find(query))
    LOG.info(matches)
    return matches


def internal_matcher(
    database, patient_obj, max_pheno_score, max_geno_score, max_results=5, score_threshold=0
):
    """Handles a query patient matching against the database of patients

    Args:
        database(pymongo.database.Database)
        patient_obj(dic): a mme formatted patient object
        max_pheno_score(float): a number between 0 and 1
        max_geno_score(float): a number between 0 and 1
        max_results(int): the maximum number of results that the server should return.
        score_threshold(float): minimum score threshold for returned results

    Returns:
        internal_match(dict): a matching object with results(list) sorted by score
    """
    json_pat = json_patient(patient_obj)
    pheno_matches = []
    pheno_m_keys = []
    geno_matches = []
    geno_m_keys = []
    matches = []

    # phenotype score can be obtained if patient has an associated phenotype (HPO or OMIM terms)
    if patient_obj.get("features") or patient_obj.get("disorders"):
        LOG.info("Matching phenotypes against database patients..")
        pheno_matches = phenomatch(
            database,
            max_pheno_score,
            patient_obj.get("features", []),
            patient_obj.get("disorders", []),
        )
        pheno_m_keys = list(pheno_matches.keys())

    # genomic score can be obtained if patient has at least one genomic feature
    if patient_obj.get("genomicFeatures") and len(patient_obj["genomicFeatures"]) > 0:
        LOG.info("Matching variants/genes against database patients..")
        geno_matches = genomatch(database, patient_obj["genomicFeatures"], max_geno_score)
        geno_m_keys = list(geno_matches.keys())

    # obtain unique list of all patient IDs returned by the 2 algorithms:
    unique_patients = list(set(pheno_m_keys + geno_m_keys))

    # create matching result objects with combined score from the 2 algorithms
    for key in unique_patients:
        pheno_score = 0
        geno_score = 0
        patient_obj = None

        if key in pheno_m_keys:
            pheno_score = pheno_matches[key]["pheno_score"]
            patient_obj = pheno_matches[key]["patient_obj"]

        if key in geno_m_keys:
            geno_score = geno_matches[key]["geno_score"]
            patient_obj = geno_matches[key]["patient_obj"]

        p_score = pheno_score + geno_score

        score = {"patient": p_score, "_genotype": geno_score, "_phenotype": pheno_score}
        match = {"patient": json_patient(patient_obj), "score": score}
        # remove matches with patient score lower than SCORE_THRESHOLD (from confif settings)
        if score["patient"] >= score_threshold:
            matches.append(match)

    # sort patient matches by patient (combined) score
    sorted_matches = sorted(matches, key=lambda k: k["score"]["patient"], reverse=True)

    # this is saved to server, regardless of the results returned by the nodes
    has_matches = bool(sorted_matches)

    internal_match = {
        "created": datetime.datetime.now(),
        "has_matches": has_matches,
        "data": {"patient": json_pat},  # description of the patient submitted
        "results": [
            {
                "node": {"id": "patientMatcher", "label": "patientMatcher server"},
                "patients": sorted_matches[:max_results],
            }
        ],
        "match_type": "internal",
    }

    # return matching object
    return internal_match


def external_matcher(database, patient, node=None):
    """Handles a query patient matching against all connected MME nodes

    Args:
        database(pymongo.database.Database)
        patient(dict) : a MME patient entity
        node(str): id of the node to search in

    Returns:
        external_match(dict): a matching object containing a list of results in 'results' field
    """
    query = {}
    if node:
        query["_id"] = node
    # get all connected nodes or ther one specified by user
    connected_nodes = list(database["nodes"].find(query))
    if len(connected_nodes) == 0:
        LOG.error("Could't find any connected MME nodes. Aborting external matching.")
        return None

    # create request headers
    headers = Headers()
    data = {"patient": json_patient(patient)}  # convert into something that follows the API specs

    # this is saved to server, regardless of the results returned by the nodes
    external_match = {
        "created": datetime.datetime.now(),
        "has_matches": False,  # it changes if a similar patient is returned by any other MME nodes
        "data": data,  # description of the patient submitted
        "results": [],
        "errors": [],
        "match_type": "external",
    }

    LOG.info("Matching patient against {} node(s)..".format(len(connected_nodes)))
    for node in connected_nodes:
        server_name = node["_id"]
        node_url = node["matching_url"]
        token = node["auth_token"]
        request_content_type = node["accepted_content"]

        headers = {
            "Content-Type": request_content_type,
            "Accept": "application/vnd.ga4gh.matchmaker.v1.0+json",
            "X-Auth-Token": token,
        }
        LOG.info('sending HTTP request to server: "{}"'.format(server_name))
        # send request and get response from server
        json_response = None
        server_return = None

        try:
            server_return = requests.request(
                method="POST", url=node_url, headers=headers, json=data
            )
            json_response = server_return.json()
        except Exception as exp:
            error = exp
            error_obj = {"node": None, "error": str(error)}
            if server_return:  # There is a response, but it's not JSON
                LOG.error("Server returned error:{}".format(error))
                error_obj["node"] = {"id": node["_id"], "label": node["label"]}
            else:  # Coudn't even send request
                LOG.error("Error while sending external match request:{}".format(error))
                error_obj["node"] = "PatientMatcher"
            external_match["errors"].append(error_obj)

        if json_response:
            LOG.info("server returns the following response: {}".format(json_response))

            result_obj = {"node": {"id": node["_id"], "label": node["label"]}, "patients": []}

            # node returned results
            if "results" in json_response:
                results = json_response["results"]
                if len(results):
                    external_match["has_matches"] = True
                    for result in results:
                        result_obj["patients"].append(result)

                    external_match["results"].append(result_obj)

            elif "query_id" in json_response:  # node is sending a query id (asynchronous response)
                #  query id must be saved in database because the node will be sending
                # a delayed request with results and this query_id as identifier
                LOG.info(
                    "Node {} is sending an async response, saving query id to server".format(
                        server_name
                    )
                )
                save_async_response(
                    database=database,
                    node_obj=node,
                    query_id=json_response["query_id"],
                    query_patient_id=patient["id"],
                )
            else:
                LOG.error("JSON response from server was:{}".format(json_response))

    return external_match


def async_match(database, response_data):
    """Creates a match object from data received from an asyc server

    Args:
        database(pymongo.database.Database)
        response_data(dict): data provided by async server in follow up request

    Returns:
        async_match(dict): a match object
    """
    # get query id you received from async server when you sent your request
    query_id = response_data["query_id"]  # can't be None, checked upstream
    # collect data saved in database when you first received the async response
    async_response = database["async_responses"].find_one({"query_id": query_id})
    patient_id = async_response.get("query_patient_id")
    patient_obj = database["patients"].find_one({"_id": patient_id})
    node = async_response.get("node")
    if not patient_obj:
        LOG.error("Couldn't find a patient with id '{} in database.".format(patient_id))
        return
    if not node:
        LOG.error("Couldn't find node '{}' in database ".format(node.get("label")))
        return
    # response key is not None, checked upstream
    resp = response_data.get("response")
    results_obj = {"node": node, "patients": resp.get("results", [])}

    async_match = {
        "created": datetime.datetime.now(),
        "has_matches": True if len(results_obj.get("patients")) > 0 else False,
        "data": {"patient": patient_obj},
        "errors": [],  # This can be modified if I know that async server returns error
        "match_type": "external",
        "async": True,
        "results": [results_obj],
    }
    return async_match


def save_async_response(database, node_obj, query_id, query_patient_id):
    """Creates an async response object in the async_responses collection.
    This object stores info about the query patient sent to an async server and the
    query_id received from it. Async server does provide match results in a
    subsequent request, along with the same query_id string, to trace it back to the
    original request initiated by patienrMatcher.

    Args:
        database(pymongo.database.Database)
        node_obj(dict): an async node object
        query_id(str): an id returned by async server in its response
        query_patient_id(str): id of internal patient matched against patients
            in async server
    """
    async_response = {
        "query_id": query_id,
        "query_patient_id": query_patient_id,
        "node": {"id": node_obj["_id"], "label": node_obj.get("label")},
        "created": datetime.datetime.now(),
    }
    database["async_responses"].insert_one(async_response)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import logging

import click
import requests
from clint.textui import progress
from flask.cli import current_app, with_appcontext
from patientMatcher.constants import PHENOTYPE_TERMS
from patientMatcher.parse.patient import EMAIL_REGEX, href_validate
from patientMatcher.utils.patient import patients

LOG = logging.getLogger(__name__)


@click.group()
def update():
    """Update patientMatcher resources"""
    pass


@update.command()
@with_appcontext
@click.option(
    "-o", "--old-href", type=click.STRING, nargs=1, required=True, help="Old contact href"
)
@click.option("-h", "--href", type=click.STRING, nargs=1, required=True, help="New contact href")
@click.option("-n", "--name", type=click.STRING, nargs=1, required=True, help="New contact name")
@click.option(
    "--institution", type=click.STRING, nargs=1, required=False, help="New contact institution"
)
def contact(old_href, href, name, institution):
    """Update contact person for a group of patients"""

    # If new contact is a simple email, add "mailto" schema
    if bool(EMAIL_REGEX.match(href)) is True and not "mailto:" in href:
        href = ":".join(["mailto", href])

    if href_validate(href) is False:
        LOG.error(
            "Provided href does not have a valid schema. Provide either a URL (http://.., https://..) or an email address (mailto:..)"
        )
        return

    database = current_app.db
    query = {"contact.href": {"$regex": old_href}}

    # Retrieving all patients matching the given old_href
    old_contact_patients = patients(database=database, match_query=query)
    # Retriving unique contacts for the above patients
    match_contacts = list(old_contact_patients.distinct("contact.href"))

    if len(match_contacts) == 0:
        click.echo(f"No patients found with contact URI '{old_href}'")
        return
    if len(match_contacts) > 1:
        click.echo(
            f"Your search for contact url '{old_href}' is returning more than one patients' contact: {match_contacts}.\nPlease restrict your search by typing a different href."
        )
        return
    # Search is returning only one contact, it's OK to use it for updating patients
    matches = list(old_contact_patients)
    new_contact = dict(href=href, name=name)
    if institution:
        new_contact["institution"] = institution

    if click.confirm(
        f"{len(matches)} patients with the old contact href '{matches[0]['contact']['href']}' will be updated with contact info:{new_contact}. Confirm?",
        abort=True,
    ):
        result = database.patients.update_many(query, {"$set": {"contact": new_contact}})
        click.echo(f"Contact information was updated for {result.modified_count} patients.")


@update.command()
@click.option("--test", help="Use this flag to test the function", is_flag=True)
def resources(test):
    """Updates HPO terms and disease ontology from the web.
    Specifically collect files from:
    http://purl.obolibrary.org/obo/hp.obo
    https://ci.monarchinitiative.org/view/hpo/job/hpo.annotations/lastSuccessfulBuild/artifact/rare-diseases/misc/phenotype_annotation.tab
    """
    files = {}
    for key, item in PHENOTYPE_TERMS.items():
        destination = item["resource_path"]
        url = item["url"]

        r = requests.get(url, stream=True)
        total_length = int(r.headers.get("content-length"))

        if test:  # read file and get its size
            files[
                key
            ] = total_length  # create an object for each downloadable file and save its length
            if total_length:
                click.echo("file {} found at the requested URL".format(key))
            continue

        with open(destination, "wb") as f:  # overwrite file
            for chunk in progress.bar(
                r.iter_content(chunk_size=1024), expected_size=(total_length / 1024) + 1
            ):
                if chunk:
                    f.write(chunk)
                    f.flush()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import patientMatcher.utils.ensembl_rest_client as ensembl_client


def entrez_to_symbol(entrez_id):
    """Convert entrez id to gene symbol

    Accepts:
        entrez_id(str) ex. "3735"

    Returns
        gene_symbol(str) ex. BRAF
    """
    client = ensembl_client.EnsemblRestApiClient()
    url = "".join([client.server, "/xrefs/name/human/", entrez_id, "?external_db=EntrezGene"])
    results = client.send_request(url)
    for entry in results:  # result is an array. First element is enough
        return entry["display_id"]


def symbol_to_ensembl(gene_symbol):
    """Convert hgnc gene symbol to Ensembl id

    Accepts:
        gene_symbol(str) ex. LIMS2

    Returns:
        ensembl_id(str) ex. ENSG00000072163
    """
    client = ensembl_client.EnsemblRestApiClient()
    # First collect all Ensembl IDs connected to a given symbol
    url = f"{client.server}/xrefs/symbol/homo_sapiens/{gene_symbol}?external_db=HGNC"

    results = client.send_request(url)
    ensembl_ids = []
    for entry in results:  # result is an array. First element is enough
        if entry["id"].startswith("ENSG") is False:
            continue
        ensembl_ids.append(entry["id"])

    if len(ensembl_ids) == 1:
        return ensembl_ids[0]

    # In case of multiple Ensembl IDs returned by the API, return only the one which has the right HGNC symbol
    for ensembl_id in ensembl_ids:
        url = f"{client.server}/xrefs/id/{ensembl_id}?all_levels=1;external_db=HGNC;content-type=application/json"
        results = client.send_request(url)
        for entry in results:
            if entry.get("display_id") == gene_symbol:
                return ensembl_id


def ensembl_to_symbol(ensembl_id):
    """Converts ensembl id to gene symbol

    Accepts:
        ensembl_id(str): an ensembl gene id. Ex: ENSG00000103591

    Returns:
        gene_symbol(str): an official gene symbol. Ex: AAGAB
    """

    client = ensembl_client.EnsemblRestApiClient()
    url = "".join([client.server, "/lookup/id/", ensembl_id])
    results = client.send_request(url)
    return results.get("display_name", None)



from patientMatcher.utils.add import load_demo_patients
from patientMatcher.utils.delete import drop_all_collections


def test_drop_all_collections(demo_data_path, database):
    """Test the functions that drop all existent collections from database before populating demo database"""
    # GIVEN a populated database
    load_demo_patients(demo_data_path, database)
    collections = database.list_collection_names()
    assert collections

    # WHEN drop_all_collections is invoked
    drop_all_collections(database)
    collections = database.list_collection_names()
    # THEN no collections should be found in database
    assert collections == []



# -*- coding: utf-8 -*-
# These classes are a simplified version of HPO and HPOIC classes
# originally developed by Orion Buske in patient-similarity (https://github.com/buske/patient-similarity)

import logging
from collections import defaultdict
from math import log

from patientMatcher.resources import path_to_hpo_terms

LOG = logging.getLogger(__name__)

ROOT = "HP:0000001"

### Code required by the HPO class ###
def get_ancestors(root, acc=None):
    if acc is None:
        acc = set()
    acc.add(root)
    for parent in root.parents:
        get_ancestors(parent, acc)
    return acc


class HPNode(object):
    """Populate an HPO object node from a list of lines in the HPO ontology file"""

    def __init__(self, hpo_lines):
        self.id = None
        self.name = None
        self.parents = set()
        self.children = set()
        self.alts = set()  # Alternative Ids
        self._parent_hps = set()
        self.obsolete = False

        if hpo_lines[0] != "[Term]":
            return

        for line in hpo_lines[1:]:
            field, value = line.split(": ", 1)
            if field == "id" and value.startswith("HP:"):
                self.id = value
            elif field == "name":
                self.name = value
            elif field == "is_a":
                hp = value.split(" !")[0]
                self._parent_hps.add(hp)
            elif field == "alt_id":
                self.alts.add(value)
            if field == "is_obsolete":
                self.obsolete = True

    def ancestors(self):
        """return all node ancestors"""
        return get_ancestors(self)

    def validate(self):
        """Make sure a node has an ID, a name and parents (if node is not root)"""
        return bool(self.id) and bool(self.name) and self.obsolete is False

    def link(self, hps):
        """Link a node objects to its parents and children on the ontology tree"""
        for hp in self._parent_hps:
            parent = hps.get(hp)
            self.parents.add(parent)
            parent.children.add(self)

    def __str__(self):
        return str(self.id)


class HPO(object):
    """Parse and HPO ontology to make it available for phenotype matching"""

    def init_app(self, app):
        """Initialize the HPO ontology when the app is launched."""
        self.hps = {}
        self._parse_ontolology()
        self.root = self.hps[ROOT]

    def _parse_ontolology(self):
        """Parse HPO ontology file, that is available under patientMatcher/resources"""
        with open(path_to_hpo_terms, encoding="utf-8") as hpofile:
            hpo_lines = []  # A group of lines containing data for an HPO term
            for line in hpofile:
                line = line.strip()
                if not line:
                    # If there is an empty line reset HPO lines
                    # since a the lines for a new HPO terms are about to start
                    hp_node = HPNode(hpo_lines)
                    # Don't take into account file header lines or obsolete terms
                    if hp_node.validate() is True:
                        self.hps[hp_node.id] = hp_node

                    # Include also alternate node IDs in hps dictionary
                    for hpid in hp_node.alts:
                        if hpid in self.hps:
                            continue
                        self.hps[hpid] = hp_node

                    hpo_lines = []  # Reset lines for next HPO term
                    continue
                hpo_lines.append(line)

        # Connect HP objects in a tree
        nodes = set(self.hps.values())

        for node in nodes:
            node.link(self.hps)

        LOG.info(f"Parsed {len(nodes)} HPO terms into HPO nodes from resource file")

    def __getitem__(self, key):
        return self.hps[key]

    def __len__(self):
        return len(set(self.hps.values()))


### End of  code required by the HPO class ###


### Code required by the HPOIC class ###
EPS = 1e-9


def _bound(p, eps=EPS):
    return min(max(p, eps), 1 - eps)


###  Class handling the information-content functionality for the HPO
class HPOIC(object):
    def init_app(self, app, hpo, diseases):
        """Initialize the HPO ontology when the app is launched."""
        LOG.info("Initializing the HPO information content")
        term_freq = self._get_term_frequencies(diseases, hpo)
        LOG.info("Total term frequency mass: {}".format(sum(term_freq.values())))
        term_ic = self._get_ics(hpo.root, term_freq)
        LOG.info("IC calculated for {}/{} terms".format(len(term_ic), len(hpo)))
        lss = self._get_link_strengths(hpo.root, term_ic)
        LOG.info("Link strength calculated for {}/{} terms".format(len(lss), len(hpo)))
        self.term_ic = term_ic
        self.lss = lss

        for term, ic in term_ic.items():
            for p in term.parents:
                assert ic >= term_ic[p] - EPS, str((term, ic, term_ic[p], p))

        LOG.info("HPO information content initialized")

    def _get_term_frequencies(self, diseases, hpo):
        """Populate term frequencies"""
        raw_freq = defaultdict(float)

        for disease in diseases.diseases.values():
            for hp_term, freq in disease.phenotype_freqs.items():
                # Try to resolve term
                try:
                    term = hpo[hp_term]
                except KeyError:
                    continue

                prevalence = 1
                freq = 1
                raw_freq[term] += freq * prevalence

        # Normalize all frequencies to sum to 1
        term_freq = {}
        total_freq = sum(raw_freq.values())
        for term in raw_freq:
            assert term not in term_freq
            term_freq[term] = _bound(raw_freq[term] / total_freq)

        return term_freq

    def _get_descendant_lookup(self, root, accum=None):
        if accum is None:
            accum = {}
        if root in accum:
            return

        descendants = set([root])
        for child in root.children:
            self._get_descendant_lookup(child, accum)
            descendants.update(accum[child])
        accum[root] = descendants
        return accum

    def _get_ics(self, root, term_freq):
        term_descendants = self._get_descendant_lookup(root)

        term_ic = {}
        for node, descendants in term_descendants.items():
            prob_mass = 0.0
            for descendant in descendants:
                prob_mass += term_freq.get(descendant, 0)

            if prob_mass > EPS:
                prob_mass = _bound(prob_mass)
                term_ic[node] = -log(prob_mass)

        return term_ic

    def _get_link_strengths(self, root, term_ic):
        lss = {}

        # P(term&parents) = P(term|parents) P(parents)
        # P(term|parents) = P(term&parents) / P(parents)
        # P(parents) = probmass(descendants)
        for term, ic in term_ic.items():
            assert term not in lss
            if term.parents:
                max_parent_ic = max([term_ic[parent] for parent in term.parents])
                ls = max(ic - max_parent_ic, 0.0)
            else:
                ls = 0.0

            lss[term] = ls

        return lss

    def information_content(self, terms):
        """Return the information content of the given terms, without backoff"""
        return sum([self.term_ic.get(term, 0) for term in terms])


### END of Code required by the HPOIC class ###



from patientMatcher.server.extensions import diseases, hpo


def test_hpo(mock_app):
    """Test the extension handling the HPO terms"""

    root_term = "HP:0000001"
    obsolete_term = "HP:0000057"

    # GIVEN the hpo extension of an app
    assert hpo
    # THEN the hpo should countain HPO terms in a dictionary
    assert isinstance(hpo.hps, dict)
    # THEN the HPO should contain ontology root term
    assert hpo[root_term]
    assert hpo[root_term].id == root_term

    # IT should also contain obsolete terms linked to non-obsolete terms
    assert hpo[obsolete_term].id != obsolete_term
    assert hpo[obsolete_term].parents
    assert hpo[obsolete_term]._parent_hps


def test_diseases(mock_app):
    """Test the extension handling disease terms"""

    # GIVEN the hpo extension of an app
    assert diseases

    # THEN the hpo should countain disease terms in a dictionary
    assert isinstance(diseases.diseases, dict)

    # GIVEN a specific term
    test_db = "OMIM"
    test_term = 201180
    # Each dictionary key should contain the expected fields
    assert diseases.diseases[(test_db, test_term)].db == test_db
    assert diseases.diseases[(test_db, test_term)].id == test_term
    assert diseases.diseases[(test_db, test_term)].phenotype_freqs



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from flask_mail import Message
from patientMatcher.cli.commands import cli
from patientMatcher.parse.patient import mme_patient
from patientMatcher.utils.ensembl_rest_client import requests


def test_appname(mock_app):
    runner = mock_app.test_cli_runner()
    result = runner.invoke(cli, ["test", "name"])
    assert result.output == "patientMatcher\n"


def test_help(mock_app):
    """Test invoking the cli with --help argument"""
    runner = mock_app.test_cli_runner()
    result = runner.invoke(cli, ["add", "client", "--help"])
    assert result.exit_code == 0


def test_run_no_args(mock_app):
    """Test invoking the cli without arguments"""
    runner = mock_app.test_cli_runner()
    result = runner.invoke(cli, [])
    assert result.exit_code == 0


def test_sendemail(mock_app, mock_mail):
    mock_app.mail = mock_mail
    mock_app.config["MAIL_USERNAME"] = "sender@email.com"

    runner = mock_app.test_cli_runner()
    # When invoking the test email command with a recipient paramrter
    result = runner.invoke(cli, ["test", "email", "--recipient", "test_user@mail.com"])

    # Make sure that mock mail send method was called and mock email is sent
    assert mock_mail._send_was_called
    assert mock_mail._message
    assert "Mail correctly sent" in result.output



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logging
from patientMatcher.parse.patient import (
    gtfeatures_to_genes_symbols,
    gtfeatures_to_variants,
    lift_variant,
)

LOG = logging.getLogger(__name__)


def match(database, gt_features, max_score):
    """Handles genotype matching algorithm

    Args:
        database(pymongo.database.Database)
        gt_features(list): a list of genomic features (objects)
        max_score(float): a number between 0 and 1

    Returns:
        matches(dict): a dictionary of patient matches with GT score
    """
    matches = {}
    n_gtfeatures = len(gt_features)

    LOG.info("\n\n###### Running genome matcher module ######")

    if n_gtfeatures > 0:
        max_feature_similarity = max_score / n_gtfeatures

        LOG.info("Query patient has {0} genotype features.".format(n_gtfeatures))
        LOG.info(
            "Each GT feature will contribute with a weight of {0} to a total GT score (max GT score is {1})".format(
                max_feature_similarity, max_score
            )
        )

        query = {}
        query_fields = []

        genes, symbols = gtfeatures_to_genes_symbols(gt_features)
        if genes:
            query_fields.append({"genomicFeatures.gene.id": {"$in": genes}})
        if symbols:
            query_fields.append({"genomicFeatures.gene._geneName": {"$in": symbols}})

        # Obtain variants and the corresponding variants in the other genome build from the genotype features
        variants = gtfeatures_to_variants(gt_features)
        if variants:
            query_fields.append({"genomicFeatures.variant": {"$in": variants}})

        if len(query_fields) > 0:
            # prepare a query that takes into account genes and variants in general (also outside genes!)
            query = {"$or": query_fields}
            LOG.info("Querying database for genomic features:{}".format(query))

            # query patients collection
            matching_patients = list(
                database["patients"].find(query)
            )  # a list of patients with genomic feature/s in one or more of the query genes
            LOG.info("Found {0} matching patients".format(len(matching_patients)))

            # assign a genetic similarity score to each of these patients
            for patient in matching_patients:
                gt_similarity = evaluate_GT_similarity(
                    gt_features, patient["genomicFeatures"], max_feature_similarity
                )
                LOG.info("GT similarity score is {}".format(gt_similarity))
                match = {
                    "patient_obj": patient,
                    "geno_score": gt_similarity,
                }
                matches[patient["_id"]] = match

    LOG.info(
        "\n\nFOUND {} patients matching patients's genomic tracts\n\n".format(
            len(matching_patients)
        )
    )
    return matches


def evaluate_GT_similarity(query_features, db_patient_features, max_feature_similarity):
    """Evaluates the genomic similarity of two patients based on genomic similarities

    Args:
        query_patient(list of dictionaries): genomic features of the query patient
        db_patient_features(list of dictionaries): genomic features of a patient in patientMatcher database
        max_feature_similarity(float): a floating point number representing the highest value allowed for a single feature

            ## Explanation: for a query patient with one feature max_similarity will be equal to MAX_GT_SCORE
               For a patient with 2 features max_similarity will be MAX_GT_SCORE/2 and so on.

    Returns:
        patient_similarity(float): the computed genetic similarity among the patients
    """

    matched_features = []
    n_feature = 0

    # loop over the query patient's features
    for feature in query_features:
        matched_features.append(0)  # score for matching of every feature is initially 0
        q_gene_id = feature["gene"]["id"]  # query feature's gene id
        q_gene_symbol = feature["gene"].get("_geneName")  # query feature's gene symbol

        # Do liftover for query variant in order to maximize perfect matching chances
        q_variant = feature.get("variant")  # query feature's variant. Not mandatory.
        lifted_q_variant = lift_variant(q_variant) if q_variant else []

        # loop over the database patient's features:
        for matching_feature in db_patient_features:
            m_gene_id = matching_feature["gene"]["id"]  # matching feature's gene id
            m_gene_symbol = matching_feature["gene"].get(
                "_geneName"
            )  # matching feature's gene symbol
            m_variant = matching_feature.get(
                "variant"
            )  # matching feature's variant. Not mandatory.

            # if variants are matching or lifted query variant matches with matched patients variant
            # ->assign max matching score
            if q_variant == m_variant or m_variant in lifted_q_variant:
                matched_features[n_feature] = max_feature_similarity

            elif q_gene_id == m_gene_id:  # matching genes
                matched_features[n_feature] = (
                    max_feature_similarity / 4
                )  # (0.25 of the max_feature_similarity)
            elif q_gene_symbol and q_gene_symbol == m_gene_symbol:
                matched_features[n_feature] = (
                    max_feature_similarity / 4
                )  # (0.25 of the max_feature_similarity)
        n_feature += 1

    features_sum = sum(matched_features)
    return features_sum



import responses
from patientMatcher.utils.variant import liftover


@responses.activate
def test_liftover_37_38():
    """Test variant liftover from GRCh37 to GRCh38"""

    # WHEN sending a liftover request with suitable coordinates in genome build 37
    chromosome = "X"
    start = 1000000
    end = 1000000

    # GIVEN a patched liftover Ensembl service
    responses.add(
        responses.GET,
        f"https://grch37.rest.ensembl.org/map/human/GRCh37/{chromosome}:{start}..{end}/GRCh38?content-type=application/json",
        json={
            "mappings": [
                {
                    "original": {
                        "assembly": "GRCh37",
                        "end": 1000000,
                        "coord_system": "chromosome",
                        "seq_region_name": "X",
                        "start": 1000000,
                        "strand": 1,
                    },
                    "mapped": {
                        "strand": 1,
                        "start": 1039265,
                        "seq_region_name": "X",
                        "end": 1039265,
                        "coord_system": "chromosome",
                        "assembly": "GRCh38",
                    },
                }
            ]
        },
        status=200,
    )

    # THEN the service should return valid mappings
    mappings = liftover("GRCh37", chromosome, start, end)

    # Each mapping should have the expected output format
    assert isinstance(mappings, list)
    assert isinstance(mappings[0], dict)
    for item in ["assembly", "seq_region_name", "start", "end"]:
        assert mappings[0]["mapped"][item]

    # And genome assembly should be GRCh38
    assert mappings[0]["mapped"]["assembly"] == "GRCh38"


@responses.activate
def test_liftover_38_37():
    """Test variant liftover from GRCh38 to GRCh37"""

    # WHEN sending a liftover request with suitable coordinates in genome build 38
    chromosome = "X"
    start = 1039265
    end = 1039265

    # GIVEN a patched liftover Ensembl service
    responses.add(
        responses.GET,
        f"https://grch37.rest.ensembl.org/map/human/GRCh38/{chromosome}:{start}..{end}/GRCh37?content-type=application/json",
        json={
            "mappings": [
                {
                    "original": {
                        "seq_region_name": "X",
                        "end": 1039265,
                        "assembly": "GRCh38",
                        "coord_system": "chromosome",
                        "strand": 1,
                        "start": 1039265,
                    },
                    "mapped": {
                        "seq_region_name": "X",
                        "end": 1000000,
                        "assembly": "GRCh37",
                        "strand": 1,
                        "coord_system": "chromosome",
                        "start": 1000000,
                    },
                },
                {
                    "original": {
                        "assembly": "GRCh38",
                        "seq_region_name": "X",
                        "end": 1039265,
                        "start": 1039265,
                        "strand": 1,
                        "coord_system": "chromosome",
                    },
                    "mapped": {
                        "seq_region_name": "HG480_HG481_PATCH",
                        "end": 1000000,
                        "assembly": "GRCh37",
                        "strand": 1,
                        "coord_system": "chromosome",
                        "start": 1000000,
                    },
                },
            ]
        },
        status=200,
    )

    # THEN the service should return valid mappings
    mappings = liftover("GRCh38", chromosome, start, end)

    # Each mapping should have the expected output format
    assert isinstance(mappings, list)
    assert isinstance(mappings[0], dict)

    for item in ["assembly", "seq_region_name", "start", "end"]:
        assert mappings[0]["mapped"][item]

    # And genome assembly should be GRCh37
    assert mappings[0]["mapped"]["assembly"] == "GRCh37"


@responses.activate
def test_liftover_MT_variant():
    """Test liftover for a mitochondrial variant from GRCh37 to GRCh38"""

    # GIVEN a mitochondrial variant
    chromosome = "MT"
    start = 7278
    end = 7278

    # GIVEN a patched liftover Ensembl service
    responses.add(
        responses.GET,
        f"https://grch37.rest.ensembl.org/map/human/GRCh37/{chromosome}:{start}..{end}/GRCh38?content-type=application/json",
        json={
            "mappings": [
                {
                    "mapped": {
                        "strand": 1,
                        "end": 7278,
                        "assembly": "GRCh38",
                        "coord_system": "chromosome",
                        "seq_region_name": "MT",
                        "start": 7278,
                    },
                    "original": {
                        "strand": 1,
                        "seq_region_name": "MT",
                        "start": 7278,
                        "end": 7278,
                        "assembly": "GRCh37",
                        "coord_system": "chromosome",
                    },
                }
            ]
        },
        status=200,
    )

    # THEN the service should return valid mappings
    mappings = liftover("GRCh37", chromosome, start, end)

    # Each mapping should have the expected output format
    assert isinstance(mappings, list)
    assert isinstance(mappings[0], dict)
    for item in ["assembly", "seq_region_name", "start", "end"]:
        assert mappings[0]["mapped"][item]

    # And genome assembly should be GRCh38
    assert mappings[0]["mapped"]["assembly"] == "GRCh38"


@responses.activate
def test_liftover_bad_request():
    """Test variant liftover with non-valid request params"""

    # WHEN sending a liftover request with non-standard chromosome (M instead of MT)
    chromosome = "M"
    start = 7278
    end = 7278

    # GIVEN a patched liftover Ensembl service
    responses.add(
        responses.GET,
        f"https://grch37.rest.ensembl.org/map/human/GRCh37/{chromosome}:{start}..{end}/GRCh38?content-type=application/json",
        json={"error": "'No slice found for M on coord_system chromosome and assembly GRCh37'"},
        status=200,
    )

    # THEN the service should not return mappings
    mappings = liftover("GRCh37", chromosome, start, end)
    assert mappings is None



# -*- coding: utf-8 -*-
import logging
import os
import sys
from pathlib import Path

import coloredlogs
from flask import Flask
from flask_mail import Mail
from patientMatcher.resources import path_to_hpo_terms, path_to_phenotype_annotations
from patientMatcher.utils.notify import TlsSMTPHandler, admins_email_format
from pymongo import MongoClient
from pymongo.errors import ConnectionFailure, OperationFailure, ServerSelectionTimeoutError

from . import extensions, views

LOG = logging.getLogger(__name__)


def available_phenotype_resources():
    """Check that necessary resources (HPO terms and phenotype annotations) were downloaded and available"""
    for filepath in [path_to_hpo_terms, path_to_phenotype_annotations]:
        if Path(path_to_hpo_terms).is_file() is False:
            return False

    return True


def configure_email_error_logging(app):
    """Setup logging of error/exceptions to email."""

    mail_handler = TlsSMTPHandler(
        mailhost=(app.config["MAIL_SERVER"], app.config["MAIL_PORT"]),
        fromaddr=app.config["MAIL_USERNAME"],
        toaddrs=app.config["ADMINS"],
        subject=f"{app.name} log error",
        credentials=(app.config["MAIL_USERNAME"], app.config["MAIL_PASSWORD"]),
    )
    mail_handler.setLevel(logging.ERROR)
    mail_handler.setFormatter(
        logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s: %(message)s " "[in %(pathname)s:%(lineno)d]"
        )
    )
    app.logger.addHandler(mail_handler)


def create_app():
    app = None
    try:
        app = Flask(__name__)
        app.config.from_envvar("PMATCHER_CONFIG")
    except:
        LOG.warning(
            "PMATCHER_CONFIG env variable not found, configuring from instance file + env vars."
        )
        app_root = os.path.abspath(__file__).split("patientMatcher")[0]

        # check if config file exists under ../instance:
        instance_path = os.path.join(app_root, "patientMatcher", "instance")
        if not os.path.isfile(os.path.join(instance_path, "config.py")):  # running app from tests
            instance_path = os.path.join(app_root, "patientMatcher", "patientMatcher", "instance")

        app = Flask(__name__, instance_path=instance_path, instance_relative_config=True)
        app.config.from_pyfile("config.py")

    if app.config.get("ADMINS"):
        app.config["ADMINS"] = admins_email_format(app.config["ADMINS"])

    current_log_level = LOG.getEffectiveLevel()
    coloredlogs.install(level="DEBUG" if app.debug else current_log_level)

    mongo_client = MongoClient(app.config["DB_URI"], serverSelectionTimeoutMS=30000)
    try:
        mongo_client.server_info()
    except (ServerSelectionTimeoutError, OperationFailure, ConnectionFailure) as err:
        LOG.warning(f"Database connection error:{err}")
        sys.exit()

    app.client = mongo_client
    db_name = app.config["DB_NAME"]
    app.db = mongo_client[db_name]
    LOG.info(f"Connecting to database '{db_name}' on {app.db}")

    # If it's not a test app and phenotype resources are missing
    # Display message and exit

    if app.config.get("TESTING") in ["False", False] and available_phenotype_resources() is False:
        LOG.error(
            "Required files hp.obo.txt and phenotype_annotation.tab.txt not found on the server. Please download them with the command 'pmatcher update resources'."
        )
        return

    extensions.hpo.init_app(app)
    extensions.diseases.init_app(app)
    extensions.hpoic.init_app(app, extensions.hpo, extensions.diseases)

    if app.config.get("MAIL_SERVER"):
        mail = Mail(app)
        app.mail = mail

        # Configure email logging of errors
        if app.debug and app.config.get("ADMINS"):
            configure_email_error_logging(app)

    app.register_blueprint(views.blueprint)

    return app



# -*- coding: utf-8 -*-
from patientMatcher.__version__ import __version__
from patientMatcher.utils.notify import (
    admins_email_format,
    html_format,
    notify_match_external,
    notify_match_internal,
)


def test_admins_email_format():
    unformatted_admins = '["foo@mail.com", "bar@mail.com"]'
    formatted_admins = admins_email_format(unformatted_admins)
    assert isinstance(formatted_admins, list)


def test_notify_match_external(mock_app, match_objs, mock_sender, mock_mail):

    match_obj = match_objs[0]  # an external match object with results
    assert match_obj["match_type"] == "external"

    with mock_app.app_context():

        # When calling the function that sends external match notifications
        notify_complete = True  # test notification of complete patient data by email
        notify_match_external(match_obj, mock_sender, mock_mail, notify_complete)

        # make sure send method was called
        assert mock_mail._send_was_called

        # and that mail object message was set correctly
        assert match_obj["data"]["patient"]["id"] in mock_mail._message.html
        assert __version__ in mock_mail._message.html


def test_notify_match_internal(database, match_objs, mock_sender, mock_mail):

    match_obj = match_objs[2]  # an internal match object with results
    assert match_obj["match_type"] == "internal"

    # insert patient used as query in database:
    assert database["patients"].find_one() is None
    assert database["patients"].insert_one({"_id": "external_patient_1"}).inserted_id

    # When calling the function that sends internal match notifications
    notify_complete = False  # test notification of partial patient data by email
    notify_match_internal(database, match_obj, mock_sender, mock_mail, notify_complete)

    # Test the function that formats the matching results to HTML:
    formatted_results = html_format(match_obj["results"])
    assert '<div style="margin-left: 0em">' in formatted_results

    # make sure send method was called
    assert mock_mail._send_was_called

    # and that mail object message was set correctly
    assert match_obj["data"]["patient"]["id"] in mock_mail._message.html
    assert __version__ in mock_mail._message.html
